/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DSMAPI_H_
#define DSMAPI_H_

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>

#include "qp_port.h"
#include "ds485.h"
#include "dsm.h"
#include "scene.h"
#include "bitset8.h"

enum DsmDone {
    DSM_DONE_NO_EVENT = 0x06
};

typedef enum {
    MOD_DEV_SENSOR_GET_VALUE           = 1,
    MOD_DEV_SENSOR_GET_TYPE            = 2,
    MOD_DEV_SENSOR_GET_COUNT           = 3,
    MOD_DEV_SENSOR_TABLE_ENTRY         = 4,
    MOD_DEV_SENSOR_EVENT               = 3,
    MOD_DEV_BINARY_INPUT_EVENT         = 4,
} DsmApiModifierDeviceSensors;

enum {
    DSM_TRANSACTION_ID  = 255,
};

enum {
    DSMAPI_COMM_LOG_UPSTREAM_MASK    = 0x01,
    DSMAPI_COMM_LOG_DOWNSTREAM_MASK  = 0x02,
};

enum {
    MOD_CLUSTER_INFO_BY_ID              = 1,
} DsmApiClusterInfoModifier;

/* analogy to ds485-core:DS485MessageOffset */
enum DS485NetlibPacketOffset {
    DS485N_PACKET_START_OFFSET     = 0,
    DS485N_PACKET_DEST_OFFSET      = 1,
    DS485N_PACKET_SRC_OFFSET       = 2,
    DS485N_PACKET_CL_OFFSET_LOW    = 3,
    DS485N_PACKET_CL_OFFSET_HIGH   = 4,
    DS485N_PACKET_DATA_OFFSET      = 5,
};

enum DSMAPIMessageOffset {
    DSMAPI_TRANSACTIONID_OFFSET = (DS485_DATA_OFFSET + 0),
    DSMAPI_MSGID_OFFSET         = (DS485_DATA_OFFSET + 1),
    DSMAPI_MODIFIER_OFFSET      = (DS485_DATA_OFFSET + 2),
    DSMAPI_ERROR_OFFSET         = (DS485_DATA_OFFSET + 3),
    DSMAPI_REQUEST_HEADER_LEN   = 3,
    DSMAPI_EVENT_HEADER_LEN     = 4,
    DSMAPI_RESPONSE_HEADER_LEN  = 4,
    DSMAPI_REQ_DATA_OFFSET      = (DS485_DATA_OFFSET + DSMAPI_REQUEST_HEADER_LEN),
    DSMAPI_EVT_DATA_OFFSET      = (DS485_DATA_OFFSET + DSMAPI_EVENT_HEADER_LEN),
    DSMAPI_RES_DATA_OFFSET      = (DS485_DATA_OFFSET + DSMAPI_RESPONSE_HEADER_LEN),
    DSMAPI_REQ_LEN_OFFSET_HI    = (DS485N_PACKET_DATA_OFFSET + 3),
    DSMAPI_REQ_LEN_OFFSET_LO    = (DS485N_PACKET_DATA_OFFSET + 4)
};

enum DSMAPILongMessageOffset {
    DSMAPI_LONG_TRANSACTIONID_OFFSET = (DS485N_PACKET_DATA_OFFSET + 0),
    DSMAPI_LONG_MSGID_OFFSET         = (DS485N_PACKET_DATA_OFFSET + 1),
    DSMAPI_LONG_MODIFIER_OFFSET      = (DS485N_PACKET_DATA_OFFSET + 2),
    DSMAPI_LONG_ERROR_OFFSET         = (DS485N_PACKET_DATA_OFFSET + 3),
    DSMAPI_LONG_RESPONSE_HEADER_LEN  = 4,
    DSMAPI_LONG_RES_DATA_OFFSET      = (DS485N_PACKET_DATA_OFFSET + DSMAPI_LONG_RESPONSE_HEADER_LEN),
};


/*..........................................................................*/
#define DsmApi_GetResData8(msg_, offset_) \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)]

#define DsmApi_GetResData16(msg_, offset_) \
    ((uint16_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+0]) << 8) | \
    ((uint16_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+1]) << 0)

#define DsmApi_GetResData32(msg_, offset_) \
    ((uint32_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+0]) << 24) | \
    ((uint32_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+1]) << 16) | \
    ((uint32_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+2]) <<  8) | \
    ((uint32_t)((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+3]) <<  0)

#define DsmApi_GetResArray8(msg_, offset_, array_, size_) \
    memcpy((array_), &((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)]), size_)

#define DsmApi_GetResError(msg_) \
    DsmApi_GetResData8(msg_, -1)

/**
 * dSM API constructor
 */
void DsmApi_ctor(bool longPowerDown);

extern QActive * const AO_DsmApi; /* "opaque" pointer to dSM API AO */

void DsmApi_SendZoneGroupActionCallSceneRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice, bool force);
void DsmApi_SendDeviceActionCallSceneRequest(uint8_t destination, uint16_t deviceId, Scene scene);
void DsmApi_SendZoneGroupActionUndoSceneNumberRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice);
void DsmApi_SendZoneGroupActionSetSceneLocalPrioRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice);
void DsmApi_SendZoneGroupActionCallSceneMinRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice);
void DsmApi_SendZoneGroupActionRegisterSceneLocalPrioRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice);
void DsmApi_SendZoneGroupActionProgOn(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice);
void DsmApi_SendDeviceActionProgOn(uint8_t destination, uint16_t deviceId);
void DsmApi_SendZoneGroupActionProgOff(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice);
void DsmApi_SendZoneGroupActionSaveScene(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice);
void DsmApi_SendZoneGroupActionSelectGroup(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice);
void DsmApi_SendZoneGroupActionLocalStop(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice);
void DsmApi_SendDeviceAccessibilityEvent(uint16_t deviceId, uint16_t zoneId, uint16_t vendorId, uint16_t productId, uint16_t functionId, uint16_t version, dsuid_t deviceDsuid, bool on);
void DsmApi_SendDevicePresenceEvent(uint16_t deviceId, bool present);
void DsmApi_SendDeviceStatisticsEvent(uint16_t deviceId, uint8_t paramOffset, uint16_t paramValue);
void DsmApi_SendDeviceActionEvent(uint16_t deviceId, uint8_t localZoneId, uint8_t groupId, uint8_t buttonNumber, uint8_t clickType);
void DsmApi_SendDeviceDirectActionEvent(uint16_t deviceId, uint8_t localZoneId, uint8_t groupId, uint16_t actionId);
void DsmApi_SendDeviceLocalActionEvent(uint16_t deviceId, uint8_t localZoneId, uint8_t localOn);
void DsmApi_SendDeviceParameterEvent(uint16_t deviceId, uint8_t paramBank, uint8_t paramOffset, uint8_t paramValue);
void DsmApi_SendZoneGroupSensorPushFromVdc(dsuid_t const *vdcDsuid, dsuid_t const *sensorDsuid,
                                           uint16_t zoneId, uint8_t groupId,
                                           uint8_t sensorType, uint16_t sensorValue,
                                           bool sensorPrecision);
void DsmApi_SendZoneGroupActionOPCDimRequest(uint8_t localZoneId, uint8_t groupId, OPCOperation operation, uint8_t channelId, uint16_t originDevice);
void DsmApi_SendDeviceActionOPCDimRequest(uint8_t destination, uint16_t deviceId, OPCOperation operation, uint8_t channelId);
void DsmApi_SendTestTransmissionQualityEvent(uint16_t deviceId, uint8_t downstreamQuality, uint16_t upstreamQuality);
void DsmApi_SendSensorListEvent(uint16_t deviceId, uint8_t sensorId, uint8_t sensorPresent, uint8_t sensorType, uint8_t lastSensor);
void DsmApi_SendSensorValueEvent(uint16_t deviceId, uint8_t sensorId, uint16_t sensorValue);
void DsmApi_SendSensorEvent(uint16_t deviceId, uint8_t eventIndex);
void DsmApi_SendBinaryInputEvent(uint16_t deviceId, uint8_t binaryInputIndex, uint8_t binaryInputType, uint8_t binaryInputState);
void DsmApi_SendUDIResponseEvent(uint16_t deviceId, uint8_t data);
void DsmApi_SendUDIInterruptEvent(uint16_t deviceId, bool highPriority);
void DsmApi_SendDeviceGiveUp(uint32_t serialNumber, uint8_t circuitId);
void DsmApi_SendShortUpstreamEvent(ShortUpstreamEvt const *e, bool crosstalk);
void DsmApi_SendLongUpstreamEvent(LongUpstreamEvt const *e, bool crosstalk);
void DsmApi_SendDeviceModelChangedEvent(uint16_t deviceAddress);
void DsmApi_SendDeviceInfoRequest(QActive * requester, uint8_t destination, uint8_t transactionId, DeviceInfoModifiers_t modifier, uint8_t localZoneId, uint16_t index);
void DsmApi_SendZoneDeviceCountRequest(QActive * requester, uint8_t destination, uint8_t transactionId, DeviceInfoModifiers_t modifier,uint8_t localZoneId);
void DsmApi_SendStatusEvent(uint8_t hardware, uint8_t temperature, uint8_t config, uint8_t ds485, uint8_t line);
void DsmApi_SendDsmDoneSetDeviceEvent(uint8_t modifier, uint8_t error, uint16_t deviceId);
void DsmApi_SendDsmDoneSetDeviceConfigEvent(uint8_t error, uint16_t deviceId, uint8_t bank, uint8_t offset);
void DsmApi_SendHeatingControllerStateEvent(uint16_t zoneId, uint8_t heatingState);

#endif /* DSMAPI_H_ */
